/*

|--------------------------------------------------------------------------|
|	RBFPRED.C: Non-Linear Time Series Predictor using a RBF Neural Network |          |
|                                                                          |
|            By Cristina & Fernando de Castro - October  2001              |
|	                                                                       |
|	V1.0 (October 2001) for DJGPP compiler                                 |
|  	V1.0.d (February 2013) adapted for DevC++ compiler with conio2 library |
|--------------------------------------------------------------------------|


		|---------------------------------------|
		|                                       |
		|   This program should be compiled     |
		|    with DevC++ compiler V 5.3.0.4     | 
		|      using conio2 library             |
		|                                       |
        |---------------------------------------|

NOTE: See Chapter 20 of  "Adaptive Filters 3rd - Simon Haykin - Prentice Hall"       
        
*/


/*****************************************************************************
*
* HEADERS:
*
*****************************************************************************/
#include<stdio.h>
#include<stdlib.h>
#include<process.h>
#include<math.h>
#include<float.h>
#include<conio2.h>
#include<ctype.h>
#include<string.h>
#include<time.h>
#include<dos.h>
#include<fcntl.h>
#include<share.h>

/*****************************************************************************
*
* PROGRAM DEFINITIONS:
*
*****************************************************************************/
#define BUFSIZE 0xFFFF /* file I/O buffer size: 0xFFFF = 64Kb */

#define SVDMAXIT 500  /* max number of iterations in SVD */

#define WMIN2MAX  1E-6 /* ratio wmin/wmax for the interpolation matrix singular values */

#define BINARYFILE  "FLOATDAT.$$$" /* binary file containing the data in the Time Series ASCII input file */

#define FRAMESIZE 0xFFFF/sizeof(float) /* FrameSize = 16383 float samples */

#define STATUS_SCREEN_TRIGGER 100 /* the divisor of the number of samples to be predicted which yields the sample interval at which the NMSE is showed in the progress status screen */

/*****************************************************************************
*
* COMPILING DIRECTIVES:
*
*****************************************************************************/

#pragma GCC diagnostic ignored "-Wwrite-strings"
#pragma GCC diagnostic ignored "-Wsequence-point"

/*****************************************************************************
*
* MACROS:
*
*****************************************************************************/


#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

static float maxarg1,maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2))

static int iminarg1,iminarg2;
#define IMIN(a,b) (iminarg1=(a),iminarg2=(b),(iminarg1) < (iminarg2) ?\
	(iminarg1) : (iminarg2))

static float sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)

static double aulimarg; /* absolute maximum upper limit */
#define AULIM(a) ( fabs(aulimarg=(a)) >= MAXFLOAT ? ( aulimarg< 0.0 ? (-MAXFLOAT) : (MAXFLOAT)) : aulimarg )


/*****************************************************************************
*
* CONSTANTS:
*
*****************************************************************************/
#define HEADERSIZE   sizeof(unsigned long)+2*sizeof(double) /* BINARYFILE Header Size */

#define MAXEXPARG 88.7 /* 88.7 = ln( 3.4x10^38): max argument value to avoid
			float overflow in function exp() */

#define MAXFLOAT 3.4E38 /* 32 bits DJGPP maximum float */

#define MINABSFLOAT 1E-45 /* 32 bits DJGPP minimum float absolute value regarded as non-zero */


/*****************************************************************************
*
* DATA TYPE STRUCTURES:
*
*****************************************************************************/
typedef struct HiddenNeuron {

float phy; /*  Hidden Neuron Output */
float NDN; /*  Normalized-to-variance RBF centers differential squared Euclidean norm */
float *t;  /*  Hidden Neuron Center Vector (size=Net->M) */

}HNEURON;



typedef struct NeuralNetwork {


unsigned K;   /* Number of Hidden Neurons = Number of RBF Gaussian Centers = Number of Process State-Vectors */
unsigned M;   /* Number of RBF Input Nodes = Prediction Order = Process State-Vectors Dimension */
float Lambda; /* RBF Regularization Parameter */
HNEURON *Hn;  /* Hidden Neurons Storage Area Pointer (size=K) */
float ss;     /* RBF Gaussian Centers Variance */
float  VarFactor;  /* ss=MaxSqEuclNormBetweenCenters*VarFactor */
float *W;      /* Output Neuron Synapses Vector (size=K) */
float y;       /* Output Neuron Output Value */
float *AuxK;   /* Temporary Auxiliary Vector (size=K) */
float **Phy;   /* Phy matrix (size= KxK) */
float **V;     /* SVD matrix V (size=KxK) */
float *w;      /* SVD vector w (size=K) */

/* -----------------------------------------------------*/

FILE *In;            /* Input Time Series Filepointer */
char *Buffer1;       /* I/O buffer #1 */
char *Buffer2;       /* I/O buffer #2 */
char *Buffer3;       /* I/O buffer #3 */
unsigned BufferSize; /* I/O Buffers #1 #2,and #3 size */
char *TSIF;          /* Time Series Input File to be read in FrameSize pieces (Frames) */
char *PASOF;         /* Predicted & Actual Series OutFile */
char *ErrOF;         /* Instant NSE & NMSE errors Output File */
char BinFile[128];   /* Temporary binary file (BINARYFILE) containing the normalized to [-1,1] samples of TSIF */
float  *Frame;       /* Sliding Frame (Frame slides on BinFile, size=FrameSize) */
float  *PredWindow;  /* Prediction Window (PredWindow slides on Frame, size=K+M) */
double A,B;          /* Normalization Parameters to range [-1,1] */
unsigned long NumElemsInFile;   /* Number of float samples in TSIF */
unsigned NumFramesInFile;       /* Number of Frames of size FrameSize in BinFile */
unsigned FrameSize;             /* Sliding Frame size */
unsigned NumElemsInXFrame;      /* Number of samples in the Extra Frame (Frame[NumFramesInFile]) */
unsigned PredWindowSize;        /* Prediction Window size (slides over Frame)   */
unsigned CurrFrame;             /* Current Frame Loaded into Memory */
unsigned SlideRangeSize;        /* Sliding Range Size of PredWindow inside the current Frame */
unsigned long LastAddrPosInFile;/* Last Addressable initial PredWindow position in BinFile */
unsigned LastAddrFrameInFile;   /* Last addressable frame in BinFile */
unsigned LastAddrFrameSize;     /* Size of the last addressable Frame in BinFile */


}NEURALNET;


typedef enum {

  NO_EVENT,           /*  " ", */
  MEMORY_ERR,         /*  "Not enough memory", */
  READ_OPEN_ERR,      /*  "I can't find the file", */
  READ_ERR,           /*  "Error reading file", */
  WRITE_OPEN_ERR,     /*  "I can't open the file", */
  WRITE_ERR,          /*  "I can't write the file", */
  BUFFER_ERR,         /*  "I can't attach buffer to file", */
  SIZE_ERR,           /*  "Specified size doesn't match data size in file", */
  MAXIT_ERR,          /*  "Exceeded max number of iterations in function", */
  LOGIC_ERR,          /*  "Logic error in function", */
  READ_NNET_ERR,      /*  "Incompatible neural network in file", */
  MTXDIM_ERR,         /*  "Invalid matrix dimensions in file", */
  PARF_ERR,           /*  "Wrong parameter file", */
  GFTD_ERR,           /*  "I can't get time & date info from file", */
  MAXFLOAT_READ_ERR,  /*  "Exceeded float maximum absolute value reading file" */
  /* ---*/
  EMPTYFILE_ERR,      /*  "Empty file" */
  WINSIZE1_ERR,       /*  "NRBF + PO must be <= than NumElemsInFile" */
  WINSIZE2_ERR        /*  "NRBF + PO must be <= than FrameSize" */

}
EVENT;


/*****************************************************************************
*
* FUNCTION PROTOTYPES:
*
*****************************************************************************/
void Quit(EVENT Evt, char *Message);
float **FloatMatCAlloc(unsigned NRows,unsigned NCols);
void FloatMatFree(float **M);
void InitNet(NEURALNET *Net, char **Argv, unsigned FrameSize, unsigned BufferSize, char *BinaryFile ); /* "constructor" */
void Help(void);
inline void Phy_O(NEURALNET *Net,float *u);
inline float MaxSqEuclNormBetweenCenters(NEURALNET *Net);
inline float DiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim);
inline float InverseOrderDiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim);
inline void InverseOrderVector(float *VInput, float *VOutput, unsigned VectorDim);
inline void SVD(float **a, int m, int n, float w[], float **v, float rv1[]);
inline float Pythag(float a, float b);
inline float **FloatMatUnitIndexDeOffSet(float **M,unsigned NRows);
inline float **FloatMatUnitIndexOffSet(float **M,unsigned NRows);
inline void PhyMatrix(NEURALNET *Net, float *u);
void NormParm(NEURALNET *Net);
void WriteNormalizedFile(NEURALNET *Net);
inline float  *SetPredWindow(NEURALNET *Net, unsigned long InitPos);
inline void SVDBackSubstitution(NEURALNET *Net,float *d);
inline void PseudoInv(NEURALNET *Net, float *u);
int GetFileTimeData(char *FileName,char *TimeData);
void RBF_TimeSeriesPredictor(NEURALNET *Net);


/*****************************************************************************
*
* GLOBAL VARIABLES:
*
*****************************************************************************/



/*****************************************************************************
*
*				MAIN():
*
*****************************************************************************/
int main(int argc, char *argv[])

{

NEURALNET Net;

/* Command Line */
if(argc!=8){Help();exit(1);}

/* InitNet : "constructor" */
InitNet(&Net,argv,FRAMESIZE,BUFSIZE,BINARYFILE);

/* RBF Neural Net based non-linear time series predictor */
RBF_TimeSeriesPredictor(&Net);

textcolor(LIGHTGRAY);
return 0;

}


/*****************************************************************************
* FUNC: void Quit(EVENT Evt, char *Message)
*
* DESC: Print Event message and exit to shell.
******************************************************************************/
void Quit(EVENT Evt, char *Message)
{

static char *Event[] = {

/*  NO_EVENT */       " ",
/*  MEMORY_ERR */     "Not enough memory",
/*  READ_OPEN_ERR */  "I can't find the file",
/*  READ_ERR */       "Error reading file",
/*  WRITE_OPEN_ERR */ "I can't open the file",
/*  WRITE_ERR */      "I can't write the file",
/*  BUFFER_ERR */     "I can't attach buffer to file",
/*  SIZE_ERR */       "Specified size doesn't match data size in file",
/*  MAXIT_ERR */      "Exceeded max number of iterations in function",
/*  LOGIC_ERR */      "Logic error in function",
/*  READ_NNET_ERR */  "Incompatible neural network in file",
/*  MTXDIM_ERR */     "Invalid matrix dimensions in file",
/*  PARF_ERR  */      "Wrong parameter file",
/*  GFTD_ERR  */      "I can't get time & date info from file",
/*  MAXFLOAT_READ_ERR */ "Exceeded float maximum absolute value reading file",
/*---------*/
/*  EMPTYFILE_ERR */   "Empty file",
/*  WINSIZE1_ERR */    "K + M must be <= NumElemsInFile",
/*  WINSIZE2_ERR */    "K + M must be <= FrameSize"
};


	system("cls");
	textcolor(LIGHTGRAY);

	if (Evt != NO_EVENT) {


		if(Message[0]){
		fprintf (stderr,"%s %s!\n", Event[Evt], Message);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}
		else{
		fprintf (stderr,"%s!\n", Event[Evt]);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}

	}
	exit (0);
}

/*****************************************************************************
* FUNC: float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
*
* DESC: Allocate memory for a NRows x NCols  float matrix.
*****************************************************************************/
float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
{

float  *x;
float  **Mx;
register unsigned i;


x = (float *)calloc((unsigned long)(NRows*NCols),sizeof(float));
if(!x)Quit(MEMORY_ERR,"");


Mx=(float  **)calloc((unsigned long)NRows,sizeof(float *));
if(!Mx)Quit(MEMORY_ERR,"");


for(i=0;i<NRows;i++){

Mx[i]=(float *)&x[i*NCols];

}

return Mx;
}

/*****************************************************************************
* FUNC: void FloatMatFree(float **M)
*
* DESC: Free memory for the float matrix M.
*****************************************************************************/
void FloatMatFree(float **M)
{

free(M[0]);
free(M);

}


/***************************************************************************
*
* FUNC: void Help(void)
*
* DESC: Help() shows the help menu.
*
****************************************************************************/
void Help(void)
{

char sep[]=" => ";

char TmpStr[128];

system("cls");
textcolor(YELLOW);

strcpy(TmpStr,"RBF Neural Network based Time Series Non-Linear Predictor V1.0.d");
gotoxy((80-strlen(TmpStr))/2,2);textcolor(YELLOW);cprintf(TmpStr);

strcpy(TmpStr,"By Fernando & Cristina de Castro - October 2001");
gotoxy((80-strlen(TmpStr))/2,4);textcolor(LIGHTBLUE);cprintf(TmpStr);

gotoxy(1,7);textcolor(LIGHTRED);cputs("Commmand Line: ");
textcolor(LIGHTCYAN);cputs("TSIF PASOF ErrOF K M R F\n\n\r");
textcolor(LIGHTCYAN);cputs("TSIF");textcolor(WHITE);cputs(sep);
textcolor(LIGHTMAGENTA);cputs("Time Series ASCII Input File\n\r");
textcolor(LIGHTCYAN);cputs("PASOF");textcolor(WHITE);cputs(sep);
textcolor(LIGHTMAGENTA);cputs("Predicted & Actual Series Output File \n\r");
textcolor(LIGHTCYAN);cputs("ErrOF");textcolor(WHITE);cputs(sep);
textcolor(LIGHTMAGENTA);cputs("Instant NSE & NMSE errors Output File\n\r");
textcolor(LIGHTCYAN);cputs("K");textcolor(WHITE);cputs(sep);
textcolor(LIGHTMAGENTA);cputs("Number of Process State-Vectors (Number of Gaussian RBF Centers)\n\r");
textcolor(LIGHTCYAN);cputs("M");textcolor(WHITE);cputs(sep);
textcolor(LIGHTMAGENTA);cputs("Prediction Order (Process State-Vectors Dimension)\n\r");
textcolor(LIGHTCYAN);cputs("R");textcolor(WHITE);cputs(sep);
textcolor(LIGHTMAGENTA);cputs("RBF Interpolation Matrix Regularization Parameter [1E-12,1E-3]\n\r");
textcolor(LIGHTCYAN);cputs("F");textcolor(WHITE);cputs(sep);
textcolor(LIGHTMAGENTA);cputs("RBFCentersVariance=Max{EuclideanNormBetweenCenters^2}*F [0.1,100]\n\r");
textcolor(LIGHTGRAY);
}

/***************************************************************************
* FUNC: inline void Phy_O(NEURALNET *Net,float *u)
*
* DESC: Determine the output value of each each hidden neuron.
****************************************************************************/
inline void Phy_O(NEURALNET *Net,float *u)
{

register unsigned k;

if(Net->ss<=0.0)Net->ss=10*MINABSFLOAT;

for(k=0;k < Net->K;k++){

Net->Hn[k].NDN=AULIM(InverseOrderDiffSquaredEuclideanNorm(u,Net->Hn[k].t,Net->M)/Net->ss);

Net->Hn[k].phy=exp(-Net->Hn[k].NDN);

}

}


/***************************************************************************
* FUNC: inline void PhyMatrix(NEURALNET *Net, float *u)
*
* DESC: Determine the interpolation matrix Phy.
****************************************************************************/
inline void PhyMatrix(NEURALNET *Net, float *u)
{

register unsigned n,k;

if(Net->ss<=0.0)Net->ss=10*MINABSFLOAT;

for(n=0;n < Net->K;n++){
for(k=0;k < Net->K;k++){

Net->Phy[n][k]=exp(-AULIM(InverseOrderDiffSquaredEuclideanNorm(u-n,Net->Hn[k].t,Net->M)/Net->ss));

if(n==k)Net->Phy[n][k]+=Net->Lambda;

}
}


}


/***************************************************************************
* FUNC: inline float MaxSqEuclNormBetweenCenters(NEURALNET *Net)
*
* DESC: Returns the square of the max distance between centers.
****************************************************************************/
inline float MaxSqEuclNormBetweenCenters(NEURALNET *Net)
{

float MaxDist2,tmp;
register unsigned i,j;

MaxDist2=10*MINABSFLOAT;;

	for(i=0;i<Net->K;i++){
	for(j=i+1;j<Net->K;j++){
		tmp=DiffSquaredEuclideanNorm(Net->Hn[i].t,Net->Hn[j].t,Net->M);
		if(tmp>MaxDist2){MaxDist2=tmp;}
	}
	}
return MaxDist2;

}




/***************************************************************************
* FUNC: inline float DiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim)
*
* DESC: Return the squared Euclidean norm of Vector0-Vector1.
****************************************************************************/
inline float DiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim)
{

register unsigned  i;

float SqEuclNorm=0;

for(i=0;i<VectorDim;i++) SqEuclNorm+=SQR(Vector0[i]-Vector1[i]);

return SqEuclNorm;


}

/***************************************************************************
* FUNC: inline float InverseOrderDiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim)
*
* DESC: Determine the squared Euclidean norm of the vector operation: InvOrder{Vector0} - Vector1
*       where InvOrder{.} is the operator which inverses the order of the vector-argument
*       components (the rightmost component is exchanged by the leftmost, and so on).
*
* NOTE: Vector Vector0 must be pointing to the last addressable position.
****************************************************************************/
inline float InverseOrderDiffSquaredEuclideanNorm(float *Vector0, float *Vector1, unsigned VectorDim)
{

register unsigned i;

float SqEuclNorm=0;

for(i=0;i<VectorDim;i++) SqEuclNorm+=SQR(Vector0[-i]-Vector1[i]);

return SqEuclNorm;

}

/***************************************************************************
* FUNC: inline void InverseOrderVector(float *VInput, float *VOutput, unsigned VectorDim)
*
* DESC: Assigns InvOrder{VInput} to VOutput, i.e., VOutput=InvOrder{VInput},
*       where InvOrder{.} is the operator which inverses the order of the vector-argument
*       components (the rightmost component is exchanged by the leftmost one, and so on).
*
* NOTE: The argument VInput must be pointing to the last addressable position.
****************************************************************************/
inline void InverseOrderVector(float *VInput, float *VOutput, unsigned VectorDim)
{

register unsigned i;

for(i=0;i<VectorDim;i++)VOutput[i]=VInput[-i];

}


/****************************************************************************
* FUNC: inline void PseudoInv(NEURALNET *Net, float *u)
*
* DESC: Determine W[]=(Phy^-1)*d[] where d[] is the desired output vector.
*
* NOTE1: Vector d[] must be pointing to the last addressable position.
*
* NOTE2: PseudoInv() uses Moore-Penrose matrix pseudo inversion via Singular
*        Value Decomposition of matrix Phy.
****************************************************************************/
inline void PseudoInv(NEURALNET *Net, float *u)
{

register unsigned i;
float wmax,wmin;

/* Singular Value Decomposition of matrix Phy */
SVD(Net->Phy,Net->K,Net->K,Net->w,Net->V,Net->AuxK);

/* kill small Singular Values to avoid ill-conditioning */
wmax=0.0;
for(i=0;i<Net->K;i++)if(Net->w[i]>wmax)wmax=Net->w[i];
wmin=wmax*WMIN2MAX;
for(i=0;i<Net->K;i++)if(Net->w[i]<wmin)Net->w[i]=0.0;


/* Back-Substitution: W[]=(Phy^-1)*d[] */
SVDBackSubstitution(Net, u);


}


/****************************************************************************
* FUNC: inline void SVDBackSubstitution(NEURALNET *Net,float *d)
*
* DESC: Determine W[]=(Phy^-1)*d[] via Singular Value Decomposition of Phy,
*       where d[] is the desired output vector.
*
* NOTE: Vector d[] must be pointing to the last addressable position.
****************************************************************************/
inline void SVDBackSubstitution(NEURALNET *Net,float *d)
{

	register unsigned jj,j,i;
	float s;

	for (j=0;j<Net->K;j++) {
		s=0.0;
		if (Net->w[j]) {
			for (i=0;i<Net->K;i++) s += Net->Phy[i][j]*d[-i];
			s /= Net->w[j];
		}
		Net->AuxK[j]=s;
	}
	for (j=0;j<Net->K;j++) {
		s=0.0;
		for (jj=0;jj<Net->K;jj++) s += Net->V[j][jj]*Net->AuxK[jj];
		Net->W[j]=s;
	}
}

/*****************************************************************************
* FUNC: inline float **FloatMatUnitIndexOffSet(float **M,unsigned NRows)
*
* DESC: Change M start index from [0][0] to [1][1].
*****************************************************************************/
inline float **FloatMatUnitIndexOffSet(float **M,unsigned NRows)
{

register unsigned i;

for(i=0;i<NRows;i++)M[i]-=1;
M-=1;

return M;
}


/*****************************************************************************
* FUNC: inline float **FloatMatUnitIndexDeOffSet(float **M,unsigned NRows)
*
* DESC: Change M start index from [1][1] to [0][0].
*****************************************************************************/
inline float **FloatMatUnitIndexDeOffSet(float **M,unsigned NRows)
{

register unsigned i;

for(i=1;i<=NRows;i++)M[i]+=1;
M+=1;

return M;
}


/*****************************************************************************
* FUNC: inline float Pythag(float a, float b)
*
* DESC: Compute (a^2+b^2)^{1/2} without overflow
*****************************************************************************/
inline float Pythag(float a, float b)
{
	float absa,absb;
	absa=fabs(a);
	absb=fabs(b);
	if (absa > absb) return absa*sqrt(1.0+SQR(absb/absa));
	else return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+SQR(absa/absb)));
}

/*****************************************************************************
* FUNC: inline void SVD(float **a, int m, int n, float w[], float **v, float rv1[])
*
* DESC: Singular value decomposition of a matrix a[mxn]. Returns matrices
*       v[nxn] and u[mxn](matrix u overwrites matrix a) and vector
*       w[n] representing the elements of the diagonal matrix w[nxn].
*       rv1[n] is an auxiliary vector.
*****************************************************************************/
inline void SVD(float **a, int m, int n, float w[], float **v, float rv1[])
{
	int flag,i,its,j,jj,k,l=0,nm=0;
	float anorm,c,f,g,h,s,scale,x,y,z;


	/* Unit Index-OffSets all storage areas */
	w--;rv1--;
	a=FloatMatUnitIndexOffSet(a,m);
	v=FloatMatUnitIndexOffSet(v,n);

	g=scale=anorm=0.0;

	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
					f=s/h;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}

		}


		w[i]=scale *g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				for (j=l;j<=m;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
					for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=FMAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	}

	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=IMIN(m,n);i>=1;i--) {
		l=i+1;
		g=w[i];
		for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			for (j=l;j<=n;j++) {
				for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
				f=(s/a[i][i])*g;
				for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else for (j=i;j<=m;j++) a[j][i]=0.0;
		++a[i][i];
	}


	for (k=n;k>=1;k--) {
		for (its=1;its<=SVDMAXIT;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if ((float)(fabs(rv1[l])+anorm) == anorm) {
					flag=0;
					break;
				}
				if ((float)(fabs(w[nm])+anorm) == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					rv1[i]=c*rv1[i];
					if ((float)(fabs(f)+anorm) == anorm) break;
					g=w[i];
					h=Pythag(f,g);
					w[i]=h;
					h=1.0/h;
					c=g*h;
					s = -f*h;
					for (j=1;j<=m;j++) {
						y=a[j][nm];
						z=a[j][i];
						a[j][nm]=y*c+z*s;
						a[j][i]=z*c-y*s;
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k] = -v[j][k];
				}
				break;
			}
			if (its >= SVDMAXIT) Quit(MAXIT_ERR,"SVD()");
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=Pythag(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=Pythag(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g = g*c-x*s;
				h=y*s;
				y *= c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=Pythag(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=c*g+s*y;
				x=c*y-s*g;
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}



	/* Unit Index-DeOffSets all storage areas */
	w++;rv1++;
	a=FloatMatUnitIndexDeOffSet(a,m);
	v=FloatMatUnitIndexDeOffSet(v,n);

}



/*****************************************************************************
* FUNC: void NormParm(NEURALNET *Net)
*
* DESC: Read the ASCII Time Series input file and determine the normalization
        parameters A and B, such that Xn=Ax+B where Xn lies in the range [-1,1].
*****************************************************************************/
void NormParm(NEURALNET *Net)
{

FILE *In;
register unsigned long i;
int test;
float v;

float maxv,minv;


maxv=-MAXFLOAT;
minv=MAXFLOAT;

/* Open Time Series input file and attach Buffer1 to it */
if((In=fopen(Net->TSIF,"rt"))==NULL)Quit(READ_OPEN_ERR,Net->TSIF);
if (setvbuf(In, Net->Buffer1, _IOFBF, Net->BufferSize) != 0)Quit(BUFFER_ERR,Net->TSIF);

i=0;
while(1){

test=fscanf(In,"%f",&v);
if(test==-1)break;
if(test!=1)Quit(READ_ERR,Net->TSIF);

if(v>maxv)maxv=v;
if(v<minv)minv=v;
i++;
}


Net->NumElemsInFile=i;
Net->A=2.0/(maxv-minv);
Net->B=(maxv+minv)/(minv-maxv);


fclose(In);

}


/*****************************************************************************
* FUNC: void WriteNormalizedFile(NEURALNET *Net)
*
* DESC: Normalize and write the Time Series input file samples to BinFile
*****************************************************************************/
void WriteNormalizedFile(NEURALNET *Net)
{

FILE *In, *Out;
register unsigned i,j;

/* Open input Time Series ASCII File to be Read and attach Buffer1 to it */
if((In=fopen(Net->TSIF,"rt"))==NULL)Quit(READ_OPEN_ERR,Net->TSIF);
if (setvbuf(In, Net->Buffer1, _IOFBF, Net->BufferSize) != 0)Quit(BUFFER_ERR,Net->TSIF);


/* Open temporary BinFile and attach Buffer2 to it */
if((Out=fopen(Net->BinFile,"wb"))==NULL)Quit(WRITE_OPEN_ERR,Net->BinFile);
if (setvbuf(Out, Net->Buffer2, _IOFBF, Net->BufferSize) != 0)Quit(BUFFER_ERR,Net->BinFile);

/* Write Parameters */
if(fwrite(&Net->NumElemsInFile,sizeof(unsigned long),1,Out)!=1)Quit(WRITE_ERR,Net->BinFile);
if(fwrite(&Net->A,sizeof(double),1,Out)!=1)Quit(WRITE_ERR,Net->BinFile);
if(fwrite(&Net->B,sizeof(double),1,Out)!=1)Quit(WRITE_ERR,Net->BinFile);


/* Read-Normalize-Write Process */
for(i=0;i<Net->NumFramesInFile;i++){

for(j=0;j<Net->FrameSize;j++){
fscanf(In,"%f",&Net->Frame[j]);
Net->Frame[j]=Net->A*Net->Frame[j]+Net->B;
}

if(fwrite((float *)Net->Frame,(unsigned)Net->FrameSize*sizeof(float),1,Out)!=1)Quit(WRITE_ERR,Net->BinFile);
}


/* If there is Extra Frame (Net->NumElemsInFile%Net->FrameSize!=0) */
if(Net->NumElemsInXFrame){
for(j=0;j<Net->NumElemsInXFrame;j++){
fscanf(In,"%f",&Net->Frame[j]);
Net->Frame[j]=Net->A*Net->Frame[j]+Net->B;
}
if(fwrite((float  *)Net->Frame,(unsigned)Net->NumElemsInXFrame*sizeof(float),1,Out)!=1)Quit(WRITE_ERR,Net->BinFile);
}


fclose(In);
fclose(Out);

}


/*****************************************************************************
* FUNC: inline float *SetPredWindow(NEURALNET *Net, unsigned long InitPos)
*
* DESC: Return Net->PredWindow[InitPos], where InitPos is the position on file BinFile.
*****************************************************************************/
inline float  *SetPredWindow(NEURALNET *Net, unsigned long InitPos)
{
static unsigned  AddressedFrame;
static unsigned AddressedPosInFrame;
static unsigned AddressedFrameSize;

AddressedFrame=(unsigned)(InitPos/Net->SlideRangeSize);
AddressedPosInFrame=(unsigned)(InitPos%Net->SlideRangeSize);

/* check end of BinFile */
if(InitPos>Net->LastAddrPosInFile) return NULL;

/* If same previous Frame return &Net->Frame[AddressedPosInFrame] */
if(AddressedFrame==Net->CurrFrame)return Net->PredWindow=&Net->Frame[AddressedPosInFrame];

/* Determine the size of the adressed frame */
if(AddressedFrame==Net->LastAddrFrameInFile)AddressedFrameSize=Net->LastAddrFrameSize;
else AddressedFrameSize=Net->FrameSize;

/* Position filepointer in the new frame and read it */
fseek(Net->In,HEADERSIZE+((unsigned long)AddressedFrame*Net->SlideRangeSize)*sizeof(float),SEEK_SET);
if(fread((float *)Net->Frame,(unsigned)AddressedFrameSize*sizeof(float),1,Net->In)!=1)Quit(READ_ERR,Net->BinFile);
Net->CurrFrame=AddressedFrame;

/* Return &Net->Frame[AddressedPosInFrame] */
return Net->PredWindow=&Net->Frame[AddressedPosInFrame];

}



/*****************************************************************************
* FUNC: void InitNet(NEURALNET *Net, char **Argv, unsigned FrameSize, unsigned BufferSize, char *BinaryFile )
*
* DESC: Init RBF Neural Network based upon command line parameters.
******************************************************************************/
void InitNet(NEURALNET *Net, char **Argv, unsigned FrameSize, unsigned BufferSize, char *BinaryFile ) /* "constructor" */
{

register unsigned i;

/* Init Net with command line parameters */
Net->TSIF=Argv[1];
Net->PASOF=Argv[2];
Net->ErrOF=Argv[3];
Net->K=atoi(Argv[4]);
Net->M=atoi(Argv[5]);
Net->PredWindowSize=Net->K+Net->M;
Net->Lambda=fabs(atof(Argv[6]));
Net->VarFactor=fabs(atof(Argv[7]));

/* Init Parameters */
Net->FrameSize=FrameSize;
if(Net->PredWindowSize>Net->FrameSize)Quit(WINSIZE2_ERR,"");
Net->BufferSize=BufferSize;
strcpy(Net->BinFile,BinaryFile);

/* Alloc Memory for Phy Matrix */
Net->Phy=FloatMatCAlloc(Net->K,Net->K);

/* Alloc Memory for SVD V Matrix */
Net->V=FloatMatCAlloc(Net->K,Net->K);

/* Alloc Memory for Slide Frame */
Net->Frame = (float  *)calloc((unsigned long)(Net->FrameSize),sizeof(float));
if(!Net->Frame)Quit(MEMORY_ERR,"");

/* Alloc memory for disk I/O speed-up buffer #1 */
Net->Buffer1=(char *)calloc(Net->BufferSize,sizeof(char));
if(!Net->Buffer1)Quit(MEMORY_ERR,"");

/* Alloc memory for disk I/O speed-up buffer #2 */
Net->Buffer2=(char *)calloc(Net->BufferSize,sizeof(char));
if(!Net->Buffer2)Quit(MEMORY_ERR,"");

/* Alloc memory for disk I/O speed-up buffer #3 */
Net->Buffer3=(char *)calloc(Net->BufferSize,sizeof(char));
if(!Net->Buffer3)Quit(MEMORY_ERR,"");

/* Alloc Memory for SVD vector w */
Net->w = (float *)calloc((unsigned long)(Net->K),sizeof(float));
if(!Net->w)Quit(MEMORY_ERR,"");

/* Alloc Memory for Output Neuron W vector */
Net->W = (float *)calloc((unsigned long)(Net->K),sizeof(float));
if(!Net->W)Quit(MEMORY_ERR,"");

/* Alloc Memory for Auxiliary vectos */
Net->AuxK = (float *)calloc((unsigned long)(Net->K),sizeof(float));
if(!Net->AuxK)Quit(MEMORY_ERR,"");


/* Alloc Memory for Hidden Neurons Storage Area  */
Net->Hn = (HNEURON *)calloc((unsigned long)(Net->K),sizeof(HNEURON));
if(!Net->Hn)Quit(MEMORY_ERR,"");

/* Alloc Memory for Hidden Neurons Center Vectors t */
for(i=0;i< Net->K;i++){
Net->Hn[i].t = (float *)calloc((unsigned long)(Net->M),sizeof(float));
if(!Net->Hn[i].t)Quit(MEMORY_ERR,"");
}


/* Read TSIF and Determine the normalization parameters A and B, such
that Xn=Ax+B where Xn lies in the range [-1,1] and X is the data set in TSIF */
NormParm(Net);

/* Determine sliding Frame / sliding PredictionWindow parameters */
Net->NumFramesInFile=(unsigned)(Net->NumElemsInFile/Net->FrameSize);
Net->NumElemsInXFrame=(unsigned)(Net->NumElemsInFile%Net->FrameSize);
if(Net->PredWindowSize>Net->NumElemsInFile)Quit(WINSIZE1_ERR,"");
Net->SlideRangeSize=Net->FrameSize-Net->PredWindowSize+1;
Net->LastAddrPosInFile=Net->NumElemsInFile-Net->PredWindowSize;
Net->LastAddrFrameInFile=(unsigned)(Net->LastAddrPosInFile/Net->SlideRangeSize);
Net->LastAddrFrameSize=(unsigned)(Net->NumElemsInFile-(Net->LastAddrFrameInFile*Net->SlideRangeSize));


/* Normalize and write TSIF data to BinFile */
WriteNormalizedFile(Net);

/* Open BinFile to be Read with the Sliding Frame and attach Buffer1 to it */
if((Net->In=fopen(Net->BinFile,"rb"))==NULL)Quit(READ_OPEN_ERR,Net->BinFile);
if (setvbuf(Net->In, Net->Buffer1, _IOFBF, Net->BufferSize) != 0)Quit(BUFFER_ERR,Net->BinFile);

/* Skip Header */
fseek(Net->In,HEADERSIZE,SEEK_SET);

/* Read Frame[0] and set Net->PredWindow to it */
if(Net->NumFramesInFile){
if(fread((float  *)Net->Frame,(unsigned)Net->FrameSize*sizeof(float),1,Net->In)!=1)Quit(READ_ERR,Net->BinFile);
}
else if(Net->NumElemsInXFrame){
if(fread((float *)Net->Frame,(unsigned)Net->NumElemsInXFrame*sizeof(float),1,Net->In)!=1)Quit(READ_ERR,Net->BinFile);
}
else Quit(EMPTYFILE_ERR,Net->BinFile);
Net->PredWindow=&Net->Frame[0];
Net->CurrFrame=0;

}


/***************************************************************************
* FUNC: RBF_TimeSeriesPredictor(NEURALNET *Net)
*
* DESC: Non-Linear Time Series Predictor via RBF Neural Network.
*
* NOTE: Centers Assignment: The last K M-dimensional process state-vectors picked up directly from the time series.
*       RBF Synapses adaption algorithm: Moore-Penrose matrix pseudo inversion via Singular Value Decomposition.
*       Variance adaption algorithm: The instant maximum squared Euclidean norm between centers scaled by VarFactor.
****************************************************************************/
void RBF_TimeSeriesPredictor(NEURALNET *Net)
{

FILE *Out1, *Out2;

register unsigned long SeriesIndex;
register unsigned k;
float *u;   /* the RBF input vector */
float NSE;  /* Instant Normalized Squared Error NSE = (u[n+1] - y)^2/(u[n+1])^2 */
float AvgNSE=0; /* Averaged NSE over the number of samples in TSIF */
float MaxNSE=0; /* Maximum NSE */
float NMSE=0; /* Normalized Mean Squared Error NMSE = ErrEnergy/PredEnergy */


unsigned long itmp;
float Obs,Obs_1=0,Pred;
float ErrEnergy=0,PredEnergy=0;

int ScreenXPos,ScreenYPos;

/* Open Predicted & Actual Series OutFile and Errors OutFile */
if((Out1=fopen(Net->PASOF,"wt"))==NULL)Quit(READ_OPEN_ERR,Net->PASOF);
if (setvbuf(Out1, Net->Buffer2, _IOFBF, Net->BufferSize) != 0)Quit(BUFFER_ERR,Net->PASOF);

if((Out2=fopen(Net->ErrOF,"wt"))==NULL)Quit(READ_OPEN_ERR,Net->ErrOF);
if (setvbuf(Out2, Net->Buffer3, _IOFBF, Net->BufferSize) != 0)Quit(BUFFER_ERR,Net->ErrOF);

SeriesIndex=0;

/* Get u[n] - the most recent M dimensional vector in PredWindow. The
 last component of u[n] is the most recent sample in PredWindow. */
u=SetPredWindow(Net,SeriesIndex)+Net->PredWindowSize-1;


/* Progress status screen init procedures */
itmp=Net->LastAddrPosInFile/STATUS_SCREEN_TRIGGER;
if(!itmp)itmp=1;


ScreenXPos=1;
ScreenYPos=wherey();

textcolor(YELLOW);

/* Prediction loop: */

while(1){

/* show progress status screen */
if(!(SeriesIndex%(itmp))||SeriesIndex==Net->LastAddrPosInFile){
gotoxy(ScreenXPos,ScreenYPos);
cprintf("Progress: %-.0f%% NMSE= %-8.3G",(float)100*SeriesIndex/Net->LastAddrPosInFile,NMSE);
}

/* Determine the RBF Centers tk starting from u[n-1] and going backwards in PredWindow */
for(k=0;k<Net->K;k++)InverseOrderVector(u-k-1,Net->Hn[k].t,Net->M);

/* Determine the square of the max distance between Centers and set it
to the Centers variance Net->ss */
Net->ss=MaxSqEuclNormBetweenCenters(Net)*Net->VarFactor;


/* Determine the interpolation matrix Phy starting from u[n-1] and going backwards in PredWindow */
PhyMatrix(Net,u-1);


/* Determine W[]=(Phy^-1)*d[] for u[n] */
PseudoInv(Net,u);


/* Determine all Hidden Neurons (RBF Centers) output (Net->Hn[k].phy) to the vector u[n] */
Phy_O(Net,u);


/* Determine RBF Neural Net output y to input vector u[n] */
Net->y=0;
for(k=0;k<Net->K;k++){
Net->y+=(Net->W[k])*(Net->Hn[k].phy);
}


SeriesIndex++;/* increment Time Series sample index */

if(kbhit())break; /* check quit by the user */
u=SetPredWindow(Net,SeriesIndex); /* slides PredWindow one position ahead in Frame */
if(u==NULL)break; /* check end of BinFile */
u+=Net->PredWindowSize-1; /* get u[n+1] */

/* Determine the normalized squared error NSE =(u[n+1]-y)^2/(u[n+1])^2 and update AvgNSE & MaxNSE */
NSE=SQR(u[0]-(Net->y))/SQR(u[0]-Net->B);
if(NSE>MaxNSE)MaxNSE=NSE;
AvgNSE+=NSE;

/* Write to Out1 the Predicted and Observed sample at SeriesIndex position */
fprintf(Out1,"%lu\t%8.6G\t%8.6G\n",SeriesIndex+Net->PredWindowSize-1,Pred=(Net->y-Net->B)/Net->A,Obs=(u[0]-Net->B)/Net->A);

/* Determine NMSE energy parameters */
if(SeriesIndex>1){
PredEnergy+=SQR(Obs-Obs_1);
ErrEnergy+=SQR(Obs-Pred);
}
else{
ErrEnergy+=SQR(Obs-Pred)/SQR(Obs);
}
Obs_1=Obs;
if(PredEnergy!=0.0)NMSE=ErrEnergy/PredEnergy;
else NMSE=ErrEnergy;


/* Write to Out2 the NSE and NMSE at SeriesIndex position */
fprintf(Out2,"%lu\t%8.6G\t%8.6G\n",SeriesIndex+Net->PredWindowSize-1,NSE,NMSE);
} /* while(1) */


/* close open files */
fclose(Out1);
fclose(Out2);
fclose(Net->In);

/* delete temp binary file */
remove(BINARYFILE);

/* write results to stdout */
textcolor(LIGHTCYAN);
cprintf("\nInput Time Series %s has %lu samples.\n\r",Net->TSIF,Net->LastAddrPosInFile+Net->PredWindowSize);
cprintf("Number of Process State-Vectors = %u\n\r",Net->K);
cprintf("Prediction Order = %u\n\r",Net->M);
cprintf("RBF Regularization Parameter = %-8.3G\n\r",Net->Lambda);
cprintf("RBF Centers Variance Factor F = %-8.3G\n\r",Net->VarFactor);
textcolor(LIGHTRED);
cprintf("Estimated Sample after %lu samples in file %s = %-8.6G\n\r",SeriesIndex+Net->PredWindowSize-1,Net->TSIF,(Net->y-Net->B)/Net->A);
cprintf("Final AvgNSE = %-8.3G\n\r",AvgNSE/=(SeriesIndex-1));
cprintf("Final MaxNSE = %-8.6G\n\r",MaxNSE);
cprintf("Final NMSE = %-8.3G\n\r",ErrEnergy/PredEnergy);

}



